function ciphertext = Encipher(plaintext, key, ciphername, mode, IV, padding)
% ciphertext = Encipher(plaintext, key, ciphername, mode, IV, paddingname)
% This function encrypts a plaintext by a cipher (default: 'AES'), for a 
% given key  and a  given mode (default: 'CBC') with a given initial vector 
% IV (default: 0) and a given padding method (default: NoPadding =
% ZeroPadding).
% The key can be a string (i.e., a normal password), which will be hashed
% by SHA-256 to get a proper key with the correct size.
% 
% This function is realized via Java Cryptography Extension.
% 
% Supported options for ciphername, mode and padding:
% ciphername: 'AES', 'AESWrap', 'ARCFOUR', 'Blowfish', 'DES', 'DESede',
%             'DESedeWrap', 'RC2', 'RC4', 'RC5'
% mode: 'ECB', 'CBC', 'PCBC', 'CFB', 'CFB8' - 'CFB128', 'OFB', 
%       'OFB8' - 'OFB128', 'CTR', 'CTS'
% padding: 'NoPadding', 'ISO10126Padding', 'OAEPPadding', 'PKCS1Padding',
%           'PKCS5Padding', 'SSL3Padding'
% For more details, see
% http://java.sun.com/javase/7/docs/technotes/guides/security/StandardNames.html#Cipher
% Note: which cipher/mode/padding method is supported depends on the
% version of JavaSE installed on your computer.
%
% Shujun Li @ www.hooklee.com 2010-2011

ciphertext = [];

if nargin<2
    disp('At least the plaintext and the key should be given as arguments!');
    return;
end
if isempty(key)
    disp('The key cannot be empty!');
    return;
end
% Check if the plaintext is valid.
p_size = size(plaintext);
p_n = numel(plaintext);
plaintext = plaintext(:)';
if (min(plaintext)<0 || max(plaintext)>255)
    disp('The plaintext is not valid!');
    return;
end
plaintext = uint8(plaintext);
% Check if the key is valid.
key = key(:)';
if (min(key)<0 || max(key)>255)
    disp('The key is not valid!');
    return;
end

if nargin<6
    padding = 'NoPadding';
    if nargin<5
        IV = zeros(1,16);
        if nargin<4
            mode = 'CBC';
            if nargin<3
                ciphername = 'AES';
            end
        end
    end
end
IV = IV(1:16)';
if (min(IV)<0 || max(IV)>255)
    disp('The initial vector is not valid!');
    return;
end
IV = uint8(IV);

% Check if the given ciphername is valid.
if ~ismember(upper(ciphername), {'AES', 'AESWRAP', 'ARCFOUR', 'BLOWFISH', ...
        'DES', 'DESEDE', 'DESEDEWRAP', 'RC2', 'RC4', 'RC5'})
    ciphername = 'AES';
end
% Set the default key size and block size.
switch(upper(ciphername))
    case {'AES', 'AESWRAP'}
        key_byte_size = 128/8;
        block_byte_size = 128/8;
    case {'DES', 'BLOWFISH'}
        key_byte_size = 56/8;
        block_byte_size = 64/8;
    case {'DESEDE', 'DESEDEWRAP'}
        key_byte_size = 168/8;
        block_byte_size = 64/8;
    case 'RC2'
        key_byte_size = 64/8;
        block_byte_size = 64/8;
    case {'ARCFOUR', 'RC4'} % Stream cipher
        key_byte_size = 128/8;
        block_byte_size = 8/8;
    case 'RC5'
        key_byte_size = 128/8;
        block_byte_size = 64/8;
end
% Check if the given mode of operation is valid.
if ~ismember(upper(mode), {'ECB', 'CBC', 'PCBC', 'CFB', 'CFB8', 'CFB16', ...
        'CFB24', 'CFB32', 'CFB40', 'CFB48', 'CFB56', 'CFB64', 'CFB72', ...
        'CFB80', 'CFB88', 'CFB96', 'CFB104', 'CFB112', 'CFB120', 'CFB128', ...
        'OFB', 'OFB8', 'OFB16', 'OFB24', 'OFB32', 'OFB40', 'OFB48', 'OFB56', ...
        'OFB64', 'OFB72', 'OFB80', 'OFB88', 'OFB96', 'OFB104', 'OFB112', ...
        'OFB120', 'OFB128', 'CTR', 'CTS'})
    ciphername = 'CBC';
end
% Check if the given padding method is valid.
if ~ismember(upper(padding), {'NOPADDING', 'ISO10126PADDING', ...
        'OAEPPADDING', 'PKCS1PADDING', 'PKCS5PADDING', 'SSL3PADDING'})
    padding = 'NoPadding';
end

% If the key is a string or a vector with less elements, hash it with 
% SHA-256 to get a key of correct size.
if (ischar(key) || numel(key)~=key_byte_size)
    Hash_Instance = java.security.MessageDigest.getInstance('SHA-256');
    Hash_Instance.update(uint8(key));
    key = typecast(Hash_Instance.digest, 'uint8')';
    key = key(1:key_byte_size);
    clear Hash_Instance;
end
% Pad the IV to have enough bytes if necessary.
if numel(IV)<block_byte_size
    IV = [IV zeros(1,block_byte_size-numel(IV))];
end
% Pad zeros to the end of the incomplete block for NoPadding/ZeroPadding
% mode.
if (strcmpi(padding, 'NoPadding') || strcmpi(padding, 'ZeroPadding'))
    if rem(p_n,block_byte_size)
        plaintext = [plaintext zeros(1,block_byte_size-rem(p_n,block_byte_size))];
    end
end

% Set the secret key for the AES instance.
Cipher_Key = javax.crypto.spec.SecretKeySpec(key, ciphername);
% Get an instance of the cipher engine and initialize it.
if ismember(upper(ciphername), {'ARCFOUR', 'RC4'}) % Stream cipher
    Cipher_Instance = javax.crypto.Cipher.getInstance(ciphername);
    Cipher_Instance.init(1, Cipher_Key); % 1 = ENCRYPT_MODE
else % Block cipher
    Cipher_Instance = javax.crypto.Cipher.getInstance([ciphername '/' mode '/' padding]);
    % % Initialize the AES instance for encryption.
    if strcmp(mode,'ECB')
        Cipher_Instance.init(1, Cipher_Key); % 1 = ENCRYPT_MODE
    else
        Cipher_Instance.init(1, Cipher_Key, javax.crypto.spec.IvParameterSpec(IV));
    end
end
% Encrypt the plaintext.
ciphertext = typecast(Cipher_Instance.doFinal(plaintext), 'uint8')';

% Transform the 1-D ciphertext back to the same dimension as the plaintext,
% if the plaintext is not padded (only for NoPadding method).
if rem(p_n,16)==0
    ciphertext = reshape(ciphertext, p_size);
end
