function dict = LZW_cipher_CPA(A, key, b, L)
% dict = LZW_cipher_CPA(A, key, b, L)
% This function implements a chosen-plaintext attack on the secure LZW
% cipher proposed by Zhou et al. in [ICME'2008, 245--248].
% The output includes the key stream used for XOR masking and the secret
% random parameters used for random permutations.
%
% Shujun Li @ www.hooklee.com 2010-2011

dict = [];

if ~exist('key','var')
    key = 'password';
end
if ~exist('b','var')
    b = 12;
elseif mod(b,2)==1
    disp('b must be an even number!');
    return;
end
N = 2^b;

if ~exist('A','var')
    plaintext_class = 'char';
    A = char([9 10 13 32:126]); % = '\t', '\n', '\r' and all other printable ASCII characters
else
    if ischar(A)
        plaintext_class = 'char';
        A = double(A); % Convert it to integers for easier handling.
    elseif (isnumeric(A) && isequal(A,fix(A)))
        plaintext_class = 'int';
        A = double(A); % Convert it to integers for easier handling.
    elseif isfloat(A)
        plaintext_class = 'float';
    else
        disp('The alphabet should be an array of chars or numbers!');
        return;
    end
end
A_n = numel(A);
if A_n<2
    disp('The size of the alphabet must be greater than 2!');
    return;
end
if N<=A_n
    disp('The size of the dictionary must be larger than the number of distinct symbols!');
    return;
end
A = A(:)'; % Make A a row vector.

t0 = tic;

% Choose A_n plaintexts in the following form:
% A(1)A(1)A(3)A(1)A(4)...A(1)A(A_n)A(1)A(2)A(2)A(4)...A(2)A(A_n)A(2)A(3)...A(A_n-1)A(A_n-1)A(A_n)A(A_n-1)A(A_n)A(A_n)
% A(2)A(2)A(4)A(2)A(5)...A(2)A(1)  A(2)A(3)A(3)A(5)...A(3)A(1)  A(3)A(4)...A(A_n)  A(A_n)  A(1)  A(A_n)  A(1)  A(1)
% ...
disp(['Generating ' num2str(A_n) ' chosen plaintexts...']);
Ss_n = A_n*(A_n-1)+2;
Ss = zeros(A_n,Ss_n);
for i=1:A_n
    Ss_i = zeros(1,Ss_n);
    AA = A([i:end 1:i-1]);
    index = 1;
    for j=1:A_n
        Ss_i(index:index+1) = AA(1);
        index = index + 2;
        if numel(AA)>=3
            AA_n_2 = 2*(numel(AA)-2) - 1;
            Ss_i(index:2:index+AA_n_2) = AA(3:end);
            Ss_i(index+1:2:index+AA_n_2) = AA(1);
            index = index + AA_n_2 + 1;
        end
        AA(1) = []; % Always remove the first element.
    end
    Ss(i,:) = Ss_i;
end
% The following code is used to avoid unnecessary computation for breaking
% short plaintexts.
if exist('L','var')
    Ss_n = min(L,Ss_n);
    Ss = Ss(:,1:Ss_n);
end
B = zeros(A_n,Ss_n);
for i=1:A_n
    disp(['Encoding-encrypting the ' iptnum2ordinal(i) ' chosen plaintext...']);
    B(i,:) = LZW_encipher(Ss(i,:), key, A, b, 0);
end
duration = toc(t0);
disp(['The time consumed on the encryption-compression process is ' seconds2time(duration) '.']);

if strcmpi(plaintext_class,'char')
    A = char(A); %#ok<NASGU>
    Ss = char(Ss);
end
save(['CPA2_b=' num2str(b) '_interim.mat'], 'B','Ss','A','key','b','N','Ss_n','A_n');

disp('Generating the equivalent dictionary...');
dict = cell(N,Ss_n);
for i=1:Ss_n
    for j=1:A_n
        dict{B(j,i)+1,i} = Ss(j,i);
    end
end

save(['CPA2_b=' num2str(b) '.mat'], 'dict','B','Ss','A','key','b');
delete(['CPA2_b=' num2str(b) '_interim.mat']);

duration = toc(t0);
disp(['The time consumed on the whole process is ' seconds2time(duration) '.']);
