function S = LZW_cracker(B, dict, A, b)
% S = LZW_cracker(B, dict, A, b)
% This function decodes and decrypts a ciphertext B based on partly broken
% key in a chosen-plaintext attack (LZW_cipher_CPA).
%
% Input arguments:
% B: the ciphertext
% dict: the dictionary obtained from the chosen-plaintext attack
% A: the source alphabet (default: 0:255 = 256 ASCII characters).
% b: defining the dictionary size 2^b (default: 12).
%
% Shujun Li @ www.hooklee.com 2010-2011

S = [];

if nargin<4
    disp('At least four input arguments are needed!');
    return;
end
if numel(B)<1
    disp('The ciphertext should contain at least one symbol!');
    return;
end
B_n = numel(B);

if ~exist('b','var')
    b = 12;
elseif mod(b,2)==1
    disp('b must be an even number!');
    return;
end
N = 2^b;
if ~exist('A','var')
    plaintext_class = 'char';
    A = char([9 10 13 32:126]); % = '\t', '\n', '\r' and all other printable ASCII characters
else
    if ischar(A)
        plaintext_class = 'char';
        A = double(A); % Convert it to integers for easier handling.
    elseif (isnumeric(A) && isequal(A,fix(A)))
        plaintext_class = 'int';
        A = double(A); % Convert it to integers for easier handling.
    elseif isfloat(A)
        plaintext_class = 'float';
    else
        disp('The alphabet should be an array of chars or numbers!');
        return;
    end
end
A_n = numel(A);
if A_n<2
    disp('The size of the alphabet must be greater than 2!');
    return;
end
if N<=A_n
    disp('The size of the dictionary must be larger than the size of the alphabet!');
    return;
end

max_S_n = min(B_n,size(dict,2));
S = cell(1,max_S_n);
disp('Decrypting and decompressing the ciphertext...');
for i=1:max_S_n
    S{i} = dict{B(i)+1,i};
    if isempty(S{i})
        if strcmpi(plaintext_class,'char')
            S{i} = char(127); % This character is never used in any real texts, so it is used here as a marker.
        else
            S{i} = NaN;
        end
    end
    switch(plaintext_class)
        case 'char'
            disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: "' char(S{i}) '".']);
        case 'int'
            disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: ' num2str(S{i},'%d ') '.']);
        otherwise % 'float'
            disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: ' num2str(S{i},'%f ') '.']);
    end
end

if strcmpi(plaintext_class,'char')
    S = cell2mat(cellfun(@char,S,'UniformOutput',false));
else
    S = cell2mat(S);
end
