function [S, X, dict] = LZW_decipher(B, key, A, b, display_level)
% [S, X, dict] = LZW_decipher(B, key, A, b, display_level)
% This function decodes and decrypts a ciphertext B with the secure LZW
% compression algorithm proposed by Zhou et al. in [ICME'2008, 245--248].
%
% Input arguments:
% B: the ciphertext
% key: the key or password, which is used by the keyed hash function (HMAC)
%       and the RC4 stream cipher involved in the encryption process
%       (default: 'password').
% A: the source alphabet (default: 0:255 = 256 ASCII characters).
% b: defining the dictionary size 2^b (default: 12).
% display_level: the level of verboseness of display: 0 = silent mode =
%                display nothing other than errors, 1 = display key
%                information (default), 2 = display detailed information
% 
% Output arguments:
% S: the plaintext
% X: the intermediate ciphertext (useful for debugging purpose)
% dict: the final LZW dictionary, which is a MXM cell array, where M =
%       2^(b/2) (useful for debugging purpose)
%
% Shujun Li @ www.hooklee.com 2010-2011

S = [];
X = [];
dict = [];

if nargin<1
    disp('At least one input argument is needed!');
    return;
end
if numel(B)<1
    disp('The ciphertext should contain at least one symbol!');
    return;
end
B_n = numel(B);

if ~exist('key','var')
    key = 'password';
end
if ~exist('b','var')
    b = 12;
elseif mod(b,2)==1
    disp('b must be an even number!');
    return;
end
b_2 = b/2;
N = 2^b;
M = 2^(b/2);
if ~exist('A','var')
    plaintext_class = 'char';
    A = 0:126; % Since '\t', '\n' and '\r' may appear, we include all characters below 127.
else
    if ischar(A)
        plaintext_class = 'char';
        A = double(A); % Convert it to integers for easier handling.
    elseif (isnumeric(A) && isequal(A,fix(A)))
        plaintext_class = 'int';
        A = double(A); % Convert it to integers for easier handling.
    elseif isfloat(A)
        plaintext_class = 'float';
    else
        disp('The alphabet should be an array of chars or numbers!');
        return;
    end
end
A_n = numel(A);
if A_n<2
    disp('The size of the alphabet must be greater than 2!');
    return;
end
if N<=A_n
    disp('The size of the dictionary must be larger than the size of the alphabet!');
    return;
end
if ~exist('display_level','var')
    display_level = 1;
end

if display_level>=1
    disp('Initializing the dictionary...');
end
dict = cell(M); % Represent the 1-D dictionary as a 2-D array to ease 2-D permutation.
for i=1:A_n
    dict = add_dict_entry(dict, A(i), key, b);
end
dict_n = A_n;

% The maximal number of bits needed is B_n*3*b, which include maximal
% B_n*2*b for generating the random permutation parameters r1, r2, c1, c2
% (all b/2-bit integers) and B_n*b bits for XOR masking,
if display_level>=1
    disp('Generating the secret bitstream used in the encryption process...');
end
PRBS = RC4_PRBS(key, B_n*3*b);
PRNS_b_2 = double(bit2num(PRBS, b_2)); % A sequence of (b/2)-bit numbers.

% Initialize the plaintext as an empty vector, which will be extended while
% the decoding/decryption process goes.
X = zeros(1,B_n);
S = [];
PRNS_i = 1; % The index of the first unused bit in the PRBS.
if display_level>=1
    disp('Decrypting and decompressing the ciphertext...');
end
for i=1:B_n
    if i>1
        c1 = PRNS_b_2(PRNS_i);
        PRNS_i = PRNS_i + 1;
        c2 = PRNS_b_2(PRNS_i);
        PRNS_i = PRNS_i + 1;
        r1 = PRNS_b_2(PRNS_i);
        PRNS_i = PRNS_i + 1;
        r2 = PRNS_b_2(PRNS_i);
        PRNS_i = PRNS_i + 1;
    end
    % XOR masking.
    K_i = bitshift(PRNS_b_2(PRNS_i),b_2) + PRNS_b_2(PRNS_i+1);
    PRNS_i = PRNS_i + 2;
    X(i) = bitxor(B(i), K_i);
    index = X(i) + 1; % 0-based index to 1-based index
    if i>1
        % Pre-permute the dictionary so that the indices of the old
        % dictionary entries match those at the encoder/encipher side.
        dict2 = permute_dict(dict, c1, c2, r1, r2);
        if isempty(dict2{index}) % Special decoding
            Ix = [I I(1)];
        else % Normal decoding
            Ix = [I dict2{index}(1)];
        end
        if dict_n<N % If the dictionary is not full.
            % Step 3-2: Randomly select an index for the new string Ix.
            [dict, index_new] = add_dict_entry(dict, Ix, key, b);
            if index_new<0
                disp('Failed to add a new entry when the dictionary is not full! This should not happen! Check the code for bug!');
                fprintf('B_i = %d\n', i);
                return;
            else
                dict_n = dict_n + 1;
                if display_level>=2
                    switch(plaintext_class)
                        case 'char'
                            disp(['  A new string of symbols was added into the dictionary: "' char(Ix) '".']);
                        case 'int'
                            disp(['  A new string of symbols was added into the dictionary: ' num2str(Ix,'%d ') '.']);
                        otherwise % 'float'
                            disp(['  A new string of symbols was added into the dictionary: ' num2str(Ix,'%f ') '.']);
                    end
                end
            end
        end
        % Step 4: Randomly permute the index array.
        dict = permute_dict(dict, c1, c2, r1, r2);
    end
    % Recover the plaintext.
    I = dict{index};
    if isempty(I)
        disp('The ciphertext or the alphabet is not valid!');
        return;
    end
    S = cat(2, S, I);
    if display_level>=2
        switch(plaintext_class)
            case 'char'
                disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: "' char(I) '".']);
            case 'int'
                disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: ' num2str(I,'%d ') '.']);
            otherwise % 'float'
                disp(['The ' iptnum2ordinal(i) ' string of symbol(s) was decoded/decrypted: ' num2str(I,'%f ') '.']);
        end
    end
end

if strcmpi(plaintext_class,'char')
    S = char(S);
elseif isequal(S,fix(S)) % integers
    if min(S(:))>=0 && max(S(:))<=255
        S = uint8(S); % Do this because in most case it is a digital image.
    end
end

if nargout>=3 % 'dict' is an output argument
    if strcmpi(plaintext_class,'char')
        for i=1:numel(dict)
            dict{i} = char(dict{i});
        end
    end
end
