function [B, K, c12_r12, X, dict, LZW0_bitsize] = LZW_encipher(S, key, A, b, display_level)
% [B, K, c12_r12, X, dict, LZW0_bitsize] = LZW_encipher(S, key, A, b, display_level)
% This function encodes and encrypts an input message S with the secure LZW
% compression algorithm proposed by Zhou et al. in [ICME'2008, 245--248].
%
% Input arguments:
% S: the input symbol sequence.
% key: the key or password, which is used by the keyed hash function (HMAC)
%       and the RC4 stream cipher involved in the encryption process
%       (default: 'password').
% A: the source alphabet (default: 32:126 = 95 ASCII characters, 0:255 =
%    Extended ASCII characters and uint8, -128:127 = int8,
%    min(S(:)):max(S(:)) = other integer types, unique(S) = float)
% b: defining the dictionary size 2^b (default: 12).
% display_level: the level of verboseness of display: 0 = silent mode =
%                display nothing other than errors, 1 = display key
%                information (default), 2 = display detailed information
% 
% Output arguments:
% B: the ciphertext
% K: the key stream used for XOR masking (useful for testing the attack)
% c12_r12: the parameters used for all random permutation operations
%          (useful for testing the attack)
% X: the intermediate ciphertext (useful for testing the attack)
% dict: the final LZW dictionary, which is a MXM cell array, where M =
%       2^(b/2) (useful for debugging purpose)
% LZW0_bitsize: the number of bits needed for variable-width LZW coding
%
% Shujun Li @ www.hooklee.com 2010-2011

B = [];
K = [];
c12_r12 = [];
X = [];
dict = [];
LZW0_bitsize = 0;

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if ~exist('key','var')
    key = 'password';
end
if ~exist('b','var')
    b = 12;
elseif mod(b,2)==1
    disp('b must be an even number!');
    return;
end
b_2 = b/2;
N = 2^b;
M = 2^(b/2);
if numel(S)<1
    disp('The plaintext should contain at least one symbol!');
    return;
end
S_class = class(S);
if ischar(S)
    plaintext_class = 'char';
    S(S>255) = ' '; % Filter some special characters greater than 255.
    S = double(S); % Convert it to double for easier handling.
elseif (isnumeric(S) && isequal(S,fix(S)))
    plaintext_class = 'int';
    S = double(S); % Convert it to integers for easier handling.
elseif isfloat(S)
    plaintext_class = 'float';
else
    disp('The plaintext should be a char or numeric array!');
    return;
end
S_min = min(S(:));
S_max = max(S(:));
if ~exist('A','var')
    switch(plaintext_class)
        case 'char'
            if (S_max<=126) % ASCII characters.
                A = 0:126; % Since '\t', '\n' and '\r' may appear, we include all characters below 127.
            else
                A = 0:255; % Extended ASCII characters.
            end
        case 'int'
            switch(S_class)
                case 'int8'
                    A = -128:127;
                case 'uint8'
                    A = 0:255;
                otherwise
                    A = S_min:S_max;
            end
        otherwise % 'float'
            A = unique(S);
    end
else
    switch(plaintext_class)
        case 'char'
            if (ischar(A) || isequal(class(A),'uint8') || ...
                (isnumeric(A) && isequal(A,fix(A)) && min(A(:))<=S_min && max(A(:))>=S_max))
                A = double(A);
            else
                disp('The alphabet does not match the input sequence of symbols!');
                return;
            end
        case {'int', 'float'}
            if isnumeric(A)
                A = double(A);
            else
                disp('Warning: change the type of the alphabet to numbers!');
            end
        otherwise % 'cell', 'struct', ...
            disp('The alphabet should be a char or numeric array!');
            return;
    end
end
A_n = numel(A);
if A_n<2
    disp('The size of the alphabet must be greater than 2!');
    return;
end
if N<=A_n
    disp('The size of the dictionary must be larger than the size of the alphabet!');
    return;
end
if ~exist('display_level','var')
    display_level = 1;
end

S_n = numel(S);
S = S(:)'; % Transform S into a row vector.

if display_level>=1
    disp('Initializing the dictionary...');
end
dict = cell(M); % Represent the 1-D dictionary as a 2-D array to ease 2-D permutation.
dict_indices = zeros(1,A_n);
for i=1:A_n
    [dict, dict_indices(i)] = add_dict_entry(dict, A(i), key, b);
end
dict_indices = dict_indices + 1; % 0-based indices to 1-based indices.
dict_n = A_n;

% The maximal number of bits needed is S_n*3*b, which include S_n*2*b for
% generating the random permutation parameters r1, r2, c1, c2 (all b/2-bit
% integers) and maximally S_n*b bits for XOR masking.
if display_level>=1
    disp('Generating the secret bitstream used in the encryption process...');
end
PRBS = RC4_PRBS(key, S_n*3*b);
PRNS_b_2 = double(bit2num(PRBS, b_2)); % A sequence of (b/2)-bit numbers.

% Initialize X, B, K and c12_r12 to be of the same size as S.
B = zeros(1,S_n);
K = zeros(1,S_n);
c12_r12 = zeros(4,S_n); % r1, r2, c1 and c2
X = zeros(1,S_n);
ciphertext_n = 0;
S_i = 1; % The index of the first to-be-coded symbol in S.
PRNS_i = 1; % The index of the first unused bit in the PRBS.
if display_level>=1
    disp('Encrypting and compressing the plaintext...');
end
while S_i<=S_n
    % Step 2: Find the longest string of symbols in the dictionary.
    index_old = -1;
    Ix = [];
    i = S_i;
    while i<=S_n
        if dict_n<N/2
            index = get_dict_index(S(S_i:i), dict, dict_indices);
        else
            index = get_dict_index(S(S_i:i), dict);
        end
        if index<0
            Ix = S(S_i:i); % This is the new entry to be added to the dictionary.
            index = index_old;
            break;
        end
        I = S(S_i:i); % This will never be empty.
        index_old = index;
        i = i + 1; % i points to the first un-encoded symbol.
    end
    if index<0
        disp('No any string is found in the dictionary! This should never happen! Check the code for bug!');
        fprintf('S_i = %d\n', S_i);
        return;
    end
    S_i = i; % This points to the first un-encoded symbol "x".
    % Step 3-1: Output the dictionary index pointing to string I.
    ciphertext_n = ciphertext_n + 1;
    X(ciphertext_n) = index;
    K(ciphertext_n) = bitshift(PRNS_b_2(PRNS_i),b_2) + PRNS_b_2(PRNS_i+1);
    PRNS_i = PRNS_i + 2;
    B(ciphertext_n) = bitxor(X(ciphertext_n), K(ciphertext_n));
    LZW0_bitsize = LZW0_bitsize + ceil(log2(dict_n));
    if display_level>=2
        switch(plaintext_class)
            case 'char'
                disp(['The ' iptnum2ordinal(ciphertext_n) ' string of symbol(s) was encoded/encrypted: "' char(I) '".']);
            case 'int'
                disp(['The ' iptnum2ordinal(ciphertext_n) ' string of symbol(s) was encoded/encrypted: ' num2str(I,'%d ') '.']);
            otherwise % 'float'
                disp(['The ' iptnum2ordinal(ciphertext_n) ' string of symbol(s) was encoded/encrypted: ' num2str(I,'%f ') '.']);
        end
    end
    if ~isempty(Ix) % This should always be non-empty except for the first round.
        if dict_n<N % If the dictionary is not full.
            % Step 3-2: Randomly select an index for the new string Ix.
            [dict, index] = add_dict_entry(dict, Ix, key, b);
            if index<0
                disp('Failed to add a new entry when the dictionary is not full! This should not happen! Check the code for bug!');
                fprintf('S_i = %d\n', S_i);
                return;
            else
                dict_n = dict_n + 1;
                if display_level>=2
                    switch(plaintext_class)
                        case 'char'
                            disp(['  A new string of symbols was added into the dictionary: "' char(Ix) '".']);
                        case 'int'
                            disp(['  A new string of symbols was added into the dictionary: ' num2str(Ix,'%d ') '.']);
                        otherwise % 'float'
                            disp(['  A new string of symbols was added into the dictionary: ' num2str(Ix,'%f ') '.']);
                    end
                end
            end
        end
    end
    % Step 4: Randomly permute the index array.
    c12_r12(:,ciphertext_n) = PRNS_b_2(PRNS_i:PRNS_i+3);
    PRNS_i = PRNS_i + 4;
    dict = permute_dict(dict, c12_r12(:,ciphertext_n));
    dict_indices = find(~cellfun(@isempty,dict))';
end
% Remove unused elements in X, B and permute_parameters.
B(ciphertext_n+1:end) = [];
K(ciphertext_n+1:end) = [];
c12_r12(:,ciphertext_n+1:end) = [];
X(ciphertext_n+1:end) = [];

if nargout>=3 % 'dict' is an output argument
    if strcmpi(plaintext_class,'char')
        for i=1:numel(dict)
            dict{i} = char(dict{i});
        end
    end
end
